package core.entities.timemanager;

import core.api.models.timemanager.tasks.TaskFieldInfo;
import core.api.models.timemanager.tasks.repeatinginfo.DeadlineStrategy;
import util.Tripel;
import util.Tupel;

import javax.persistence.*;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.*;

@Entity
@Table(name = "tasks")
public class Task {

    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    private long taskID;

    @ManyToOne
    @JoinColumn(name = "taskgroup_id")
    private Taskgroup taskgroup;
    private String taskName;
    private LocalDate startDate;
    private LocalDate deadline;
    private int eta;
    private boolean finished;
    private boolean finishable;
    @OneToMany(mappedBy = "task", cascade = CascadeType.ALL, fetch = FetchType.EAGER, orphanRemoval = true)
    private List<AbstractSchedule> basicTaskSchedules;
    private int workTime;
    @OneToOne(mappedBy = "task", cascade = CascadeType.ALL, orphanRemoval = true)
    private TaskSerieItem taskSerieItem;

    @OneToMany(mappedBy = "parent", cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.EAGER)
    private Set<Task> subtasks;

    @ManyToOne
    @JoinColumn(name = "parent")
    private Task parent;

    public Task() {
        this.basicTaskSchedules = new ArrayList<>();
    }

    public Task(Taskgroup taskgroup, TaskFieldInfo taskFieldInfo) {
        this.taskgroup = taskgroup;
        this.taskName = taskFieldInfo.getTaskName();
        this.startDate = taskFieldInfo.getStartDate();
        this.deadline = taskFieldInfo.getDeadline();
        this.eta = taskFieldInfo.getEta();
        this.finished = false;
        this.finishable = taskFieldInfo.isFinishable();
    }

    public Tripel<Task, Collection<Task>, Collection<AbstractSchedule>> cloneTask() {
       Collection<Task> clonedTasks = new ArrayList<>();
       Collection<AbstractSchedule> clonedSchedules = new ArrayList<>();

       Task clonedTask = new Task();
       clonedTasks.add(clonedTask);

       clonedTask.setTaskgroup(this.getTaskgroup());
       clonedTask.setTaskName(this.taskName);
       clonedTask.setEta(this.eta);
       clonedTask.setFinished(this.finished);
       clonedTask.setFinishable(this.finishable);
       clonedTask.setStartDate(this.startDate);
       clonedTask.setDeadline(this.deadline);

       for(AbstractSchedule abstractSchedule : this.basicTaskSchedules) {
           AbstractSchedule clonedSchedule = abstractSchedule.cloneSchedule();
           clonedSchedules.add(clonedSchedule);
           clonedTask.getBasicTaskSchedules().clear();
           clonedTask.getBasicTaskSchedules().add(clonedSchedule);
       }

       Set<Task> clonedSubtasks = new HashSet<>();
       for(Task task : this.subtasks) {
           Tripel<Task, Collection<Task>, Collection<AbstractSchedule>> clonedSubtask = task.cloneTask();
           clonedSubtask.getValue00().setParent(clonedTask);
           clonedSubtasks.add(clonedSubtask.getValue00());

           clonedTasks.addAll(clonedSubtask.getValue01());
           clonedSchedules.addAll(clonedSubtask.getValue02());
       }
       clonedTask.setSubtasks(clonedSubtasks);
       return new Tripel<>(clonedTask, clonedTasks, clonedSchedules);
    }

    public void shiftTask(long offset) {
       this.setStartDate(this.getStartDate().plusDays(offset));
       this.setDeadline(this.getDeadline().plusDays(offset));

       for(AbstractSchedule abstractSchedule : this.basicTaskSchedules) {
           abstractSchedule.shiftSchedule(offset);
       }

       for(Task subtask: this.subtasks) {
           subtask.shiftTask(offset);
       }
    }


    public long getTaskID() {
        return taskID;
    }

    public Taskgroup getTaskgroup() {
        return taskgroup;
    }

    public void setTaskgroup(Taskgroup taskgroup) {
        this.taskgroup = taskgroup;
    }

    public String getTaskName() {
        return taskName;
    }

    public void setTaskName(String taskName) {
        this.taskName = taskName;
    }

    public LocalDate getStartDate() {
        return startDate;
    }

    public void setStartDate(LocalDate startDate) {
        this.startDate = startDate;
    }

    public LocalDate getDeadline() {
        return deadline;
    }

    public void setDeadline(LocalDate deadline) {
        this.deadline = deadline;
    }

    public int getEta() {
        return eta;
    }

    public void setEta(int eta) {
        this.eta = eta;
    }

    public void finish() {
        this.finished = !this.finished;
    }

    public boolean isFinished() {
        return finished;
    }

    public void setFinished(boolean finished) {
        this.finished = finished;
    }

    public int getWorkTime() {
        return workTime;
    }

    public void setWorkTime(int workTime) {
        this.workTime = workTime;
    }

    public void addWorkTime(int workTime) {
        this.workTime += workTime;
    }

    public void setTaskID(long taskID) {
        this.taskID = taskID;
    }

    public TaskSerieItem getTaskSerieItem() {
        return taskSerieItem;
    }

    public void setTaskSerieItem(TaskSerieItem taskSerieItem) {
        this.taskSerieItem = taskSerieItem;
    }

    public List<AbstractSchedule> getBasicTaskSchedules() {
        if(basicTaskSchedules == null) {
            return new ArrayList<>();
        }
        return basicTaskSchedules;
    }

    public void setBasicTaskSchedules(List<AbstractSchedule> basicTaskSchedules) {
        this.basicTaskSchedules = basicTaskSchedules;
    }

    public void addSubtask(Task subtask) {
        subtask.setParent(this);
        this.subtasks.add(subtask);
    }

    public Set<Task> getSubtasks() {
        return subtasks;
    }

    public void setSubtasks(Set<Task> subtasks) {
        this.subtasks = subtasks;
    }

    public Task getParent() {
        return parent;
    }

    public void setParent(Task parent) {
        this.parent = parent;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Task task = (Task) o;
        return taskID == task.taskID;
    }

    @Override
    public int hashCode() {
        return Objects.hash(taskID, taskgroup, taskName, startDate, deadline, eta, finished, workTime);
    }

    @Override
    public String toString() {
        return "Task{" +
                "taskID=" + taskID +
                ", taskgroup=" + taskgroup +
                ", taskName='" + taskName + '\'' +
                ", startDate=" + startDate +
                ", deadline=" + deadline +
                ", eta=" + eta +
                ", finished=" + finished +
                ", workTime=" + workTime +
                '}';
    }

    public boolean hasActiveSchedule() {
        for (AbstractSchedule schedule : getBasicTaskSchedules()) {
            if (schedule.isActive()) {
                return true;
            }
        }
        return false;
    }

    public List<AbstractSchedule> getStartableSchedules() {
        List<AbstractSchedule> abstractSchedules = new ArrayList<>();
        for(AbstractSchedule schedule : getBasicTaskSchedules()) {
            if(schedule.isStartable()) {
                abstractSchedules.add(schedule);
            }
        }
        return abstractSchedules;
    }

    public boolean isFinishable() {
        return finishable;
    }

    public void setFinishable(boolean finishable) {
        this.finishable = finishable;
    }

    public int calcOverallActivityInfo(LocalDate date) {
        int activeMinutes = 0;
        for(AbstractSchedule schedule : getBasicTaskSchedules()) {
            if(schedule.isCompleted() && schedule.getStartTime().toLocalDate().isEqual(date)) {
                activeMinutes += schedule.calcActiveMinutes();
            }
        }
        return activeMinutes;
    }

    public boolean hasPlannedSchedules() {
        for(AbstractSchedule schedule : getBasicTaskSchedules()) {
            if(!schedule.isCompleted()) {
                return true;
            }
        }
        return false;
    }

    public void increaseWorkTime(long minutes) {
        this.workTime += (int) minutes;
    }
}
